# -*- coding: utf-8 -*-
from __future__ import absolute_import

# Third Party Stuff
from django.conf import settings as django_settings
from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client as TwilioRestClient

# Local
from .base import BaseBackend


class TwilioBackend(BaseBackend):
    def __init__(self, **options):
        super(TwilioBackend, self).__init__(**options)
        # Lower case it just to be sure
        options = {key.lower(): value for key, value in options.items()}
        self._sid = options.get("sid", None)
        self._secret = options.get("secret", None)  # auth_token
        self._from = options.get("from", None)

        self.client = TwilioRestClient(self._sid, self._secret)
        self.exception_class = TwilioRestException

    def send_sms(self, number, message):
        self.client.messages.create(to=number, body=message, from_=self._from)

    def send_bulk_sms(self, numbers, message):
        for number in numbers:
            self.send_sms(to=number, body=message, from_=self._from)


class TwilioSandboxBackend(BaseBackend):
    def __init__(self, **options):
        super(TwilioSandboxBackend, self).__init__(**options)
        # Lower case it just to be sure
        options = {key.lower(): value for key, value in options.items()}
        self._sid = options.get("sid", None)
        self._secret = options.get("secret", None)  # auth_token
        self._from = options.get("from", None)
        self._token = django_settings.PHONE_VERIFICATION.get("TWILIO_SANDBOX_TOKEN")

        self.client = TwilioRestClient(self._sid, self._secret)
        self.exception_class = TwilioRestException

    def send_sms(self, number, message):
        self.client.messages.create(to=number, body=message, from_=self._from)

    def send_bulk_sms(self, numbers, message):
        for number in numbers:
            self.send_sms(to=number, body=message, from_=self._from)

    def generate_token(self):
        """
        Returns an fixed token
        """
        return self._token

    def create_temporary_token(self, number):
        """
        Creates a temporary token inside the cache, this holds the phone number
        as value, so that we can later check if everything is correct.

        :param number: Number of recipient

        :return token: string of SHA token
        """
        return self.generate_token()

    def validate_token(self, otp, phone_number):
        return self.VALID



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# Third party stuff
import json
import time

import pytest
from django.apps import apps
from django.urls import reverse

from . import factories as f

pytestmark = pytest.mark.django_db

OTP = "123456"
PHONE_NUMBER = "+13478379634"
SESSION_CODE = "phone-auth-session-code"


def test_phone_registration_sends_message(client, mocker):
    url = reverse("phone-register")
    phone_number = PHONE_NUMBER
    data = {"phone_number": phone_number}
    twilio_api = mocker.patch(
        "phone_verify.services.PhoneVerificationService.send_verification"
    )

    response = client.post(url, data)

    assert response.status_code == 200
    assert twilio_api.called
    assert "session_code" in response.data
    SMSVerification = apps.get_model("phone_verify", "SMSVerification")
    assert SMSVerification.objects.get(
        session_code=response.data["session_code"], phone_number=phone_number
    )


def test_otp_session_code_verification_api(client):
    f.create_verification(otp=OTP, phone_number=PHONE_NUMBER, session_code=SESSION_CODE)
    url = reverse("phone-verify")
    data = {"phone_number": PHONE_NUMBER, "otp": OTP, "session_code": SESSION_CODE}
    response = client.json.post(url, data=data)
    assert response.status_code == 200
    assert response.data["message"] == "OTP is valid."


def test_phone_verification_with_incomplete_payload(client):
    f.create_verification(otp=OTP, phone_number=PHONE_NUMBER, session_code=SESSION_CODE)
    url = reverse("phone-verify")
    data = {"phone_number": PHONE_NUMBER}
    response = client.json.post(url, data=data)
    assert response.status_code == 400
    response_data = json.loads(json.dumps(response.data))
    assert response_data["session_code"][0] == "This field is required."
    assert response_data["otp"][0] == "This field is required."

    data = {"otp": OTP}
    response = client.json.post(url, data=data)
    assert response.status_code == 400
    response_data = json.loads(json.dumps(response.data))
    assert response_data["phone_number"][0] == "This field is required."


def test_phone_verification_with_incorrect_payload(client):
    f.create_verification(otp=OTP, phone_number=PHONE_NUMBER, session_code=SESSION_CODE)
    url = reverse("phone-verify")
    # Payload with wrong session code
    data = {
        "phone_number": PHONE_NUMBER,
        "otp": OTP,
        "session_code": "wrong-session-code",
    }
    response = client.json.post(url, data=data)
    response_data = json.loads(json.dumps(response.data))
    assert response.status_code == 400
    assert response_data["non_field_errors"][0] == "Session Code mis-match"

    # Payload with wrong OTP
    data = {"phone_number": PHONE_NUMBER, "otp": "999999", "session_code": SESSION_CODE}
    response = client.json.post(url, data=data)
    assert response.status_code == 400
    response_data = json.loads(json.dumps(response.data))
    assert response_data["non_field_errors"][0] == "OTP is not valid"

    # Payload with incorrect phone_number
    data = {"phone_number": "+13478379632", "otp": OTP, "session_code": SESSION_CODE}
    response = client.json.post(url, data=data)
    assert response.status_code == 400
    response_data = json.loads(json.dumps(response.data))
    assert response_data["non_field_errors"][0] == "OTP is not valid"


def test_otp_expired(client):
    f.create_verification(otp=OTP, phone_number=PHONE_NUMBER, session_code=SESSION_CODE)
    time.sleep(2)
    url = reverse("phone-verify")
    data = {"phone_number": PHONE_NUMBER, "otp": OTP, "session_code": SESSION_CODE}
    response = client.json.post(url, data=data)
    assert response.status_code == 400
    response_data = json.loads(json.dumps(response.data))
    assert response_data["non_field_errors"][0] == "OTP has expired"



# -*- coding: utf-8 -*-

# Third party
from django.conf import settings
from django.utils.module_loading import import_string

DEFAULT_SERVICE = "phone_verify.backends.base.BaseBackend"

backend = None


def get_sms_backend(phone_number):
    if not backend:
        backend_import = DEFAULT_SERVICE

        if settings.PHONE_VERIFICATION.get("BACKEND", None):
            backend_import = settings.PHONE_VERIFICATION["BACKEND"]

        backend_cls = import_string(backend_import)
        return backend_cls(**settings.PHONE_VERIFICATION["OPTIONS"])



# -*- coding: utf-8 -*-

# Third Party Stuff
import pytest

# phone_verify Stuff
from phone_verify.services import PhoneVerificationService

pytestmark = pytest.mark.django_db


def test_message_generation_and_sending_service(client, mocker):
    service = PhoneVerificationService(phone_number="+13478379634")
    twilio_api = mocker.patch("phone_verify.backends.twilio.TwilioBackend.send_sms")
    service.send_verification("+13478379634", "123456")

    assert twilio_api.called



# -*- coding: utf-8 -*-

# Third Party Stuff
import jwt
from django.conf import settings as django_settings
from django.utils import timezone
from django.utils.crypto import get_random_string

from ..models import SMSVerification

DEFAULT_TOKEN_LENGTH = 6


class BaseBackend(object):
    VALID = 0
    INVALID = 1
    EXPIRED = 2

    def __init__(self, **settings):
        self.exception_class = None

    def send_sms(self, numbers, message):
        raise NotImplementedError()

    def send_bulk_sms(self, numbers, message):
        raise NotImplementedError()

    @classmethod
    def generate_token(cls):
        """
        Returns an unique random token
        """
        token_length = django_settings.PHONE_VERIFICATION.get(
            "TOKEN_LENGTH", DEFAULT_TOKEN_LENGTH
        )
        return get_random_string(token_length, allowed_chars="0123456789")

    @classmethod
    def generate_session_token(cls, token, phone_number):
        """
        Returns an unique random token for a particular device
        """
        data = {"device_%s_session_code" % (phone_number): token}
        return jwt.encode(data, django_settings.SECRET_KEY).decode()

    @classmethod
    def token_expired(cls, stored_verification):
        time_difference = timezone.now() - stored_verification.created_at
        if time_difference.seconds > django_settings.PHONE_VERIFICATION.get(
            "OTP_EXPIRATION_TIME"
        ):
            return True
        return False

    def create_otp_and_session_token(self, number):
        """
        Creates a temporary otp (OTP) inside the cache, this holds the phone number
        as value, so that we can later check if everything is correct.

        It also generates a session_code for the client to send in subsequent requests.

        :param number: Number of recipient

        :return otp: string of sha otp
        :return session_code: string of session_code
        """
        otp = self.generate_token()
        session_code = self.generate_session_token(otp, number)

        # Delete old OTPs for phone_number if already exists
        SMSVerification.objects.filter(phone_number=number).delete()

        # Default otp generated of 6 digits
        SMSVerification.objects.create(
            phone_number=number, otp=otp, session_code=session_code
        )
        return otp, session_code

    def validate_token(self, otp, phone_number):
        stored_verification = SMSVerification.objects.filter(
            otp=otp, phone_number=phone_number
        ).first()

        # check otp exists
        if stored_verification is None:
            return stored_verification, self.INVALID

        # check otp is not expired
        if self.token_expired(stored_verification):
            return stored_verification, self.EXPIRED
        return stored_verification, self.VALID



# -*- coding: utf-8 -*-

# Standard Library
import logging

# Third Party Stuff
from django.utils.translation import ugettext as _
from phonenumber_field.serializerfields import PhoneNumberField
from rest_framework import serializers

# Phone Auth Stuff
from .backends import get_sms_backend

logger = logging.getLogger(__name__)


class PhoneSerializer(serializers.Serializer):
    phone_number = PhoneNumberField()


class SMSVerificationSerializer(serializers.Serializer):
    phone_number = PhoneNumberField(required=True)
    session_code = serializers.CharField(required=True)
    otp = serializers.CharField(required=True)

    def validate(self, attrs):
        attrs = super().validate(attrs)
        phone_number = attrs.get("phone_number", None)
        otp, session_code = attrs.get("otp", None), attrs.get("session_code", None)
        backend = get_sms_backend(phone_number=phone_number)
        verification, token_validatation = backend.validate_token(
            otp=otp, phone_number=phone_number
        )

        if verification is None:
            raise serializers.ValidationError(_("OTP is not valid"))
        elif not verification.session_code == session_code:
            raise serializers.ValidationError(_("Session Code mis-match"))
        elif token_validatation == backend.EXPIRED:
            raise serializers.ValidationError(_("OTP has expired"))
        return attrs



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# Third Party Stuff
from django.contrib import admin

from .models import SMSVerification


@admin.register(SMSVerification)
class SMSVerificationAdmin(admin.ModelAdmin):
    list_display = ("id", "otp", "phone_number", "created_at")
    search_fields = ("phone_number",)
    ordering = ("phone_number",)
    readonly_fields = (
        "otp",
        "phone_number",
        "session_code",
        "created_at",
        "modified_at",
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

# Third Party Stuff
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

# phone_verify stuff
from .backends import get_sms_backend

logger = logging.getLogger(__name__)

# iOS uses "session code" to parse the OTP and support copying on clipboard.
DEFAULT_MESSAGE = "Welcome to {app}, use session code {otp} for authentication."
DEFAULT_APP_NAME = "Phone Verify"


class PhoneVerificationService(object):
    try:
        phone_settings = settings.PHONE_VERIFICATION
    except AttributeError:
        raise ImproperlyConfigured("Please define PHONE_VERIFICATION in settings")

    verification_message = phone_settings.get("MESSAGE", DEFAULT_MESSAGE)

    def __init__(self, phone_number, backend=None):
        if backend is None:
            self.backend = get_sms_backend(phone_number=phone_number)

    def send_verification(self, number, otp):
        """
        Send a verification text to the given number to verify.

        :param number: the phone number of recipient.
        """
        message = self._generate_message(otp)

        self.backend.send_sms(number, message)

    def _generate_message(self, otp):
        return self.verification_message.format(
            app=settings.PHONE_VERIFICATION.get("APP_NAME", DEFAULT_APP_NAME), otp=otp
        )


def send_otp_and_generate_session_code(phone_number):
    sms_backend = get_sms_backend(phone_number)
    otp, session_code = sms_backend.create_otp_and_session_token(phone_number)
    service = PhoneVerificationService(phone_number=phone_number)
    try:
        service.send_verification(phone_number, otp)
    except service.backend.exception_class as exc:
        logger.error(
            "Error in sending verification code to {phone_number}: "
            "{error}".format(phone_number=phone_number, error=exc)
        )
    return session_code
